from search_space.errors import DetectedRuntimeDependency
from . import VisitorLayer
from search_space.utils import visitor
from search_space.spaces.asts import naturals_values
from search_space.spaces.asts.naturals_values import NaturalValuesNode


class NaturalAstVisitor:
    def get_value(self, node, current_index=[], context=None):
        self.current_index, self._context = current_index, context
        return self.visit(node)

    @property
    def context(self):
        try:
            if self._context != None:
                return self._context
        except AttributeError:
            pass

        raise DetectedRuntimeDependency()

    @visitor.on("node")
    def visit(self, node):
        pass

    #################################################################
    #                                                               #
    #                 Arithmetic Visit                              #
    #                                                               #
    #################################################################

    @visitor.when(naturals_values.ModOp)
    def visit(self, node: naturals_values.ModOp):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a % b

    @visitor.when(naturals_values.AddOp)
    def visit(self, node: naturals_values.AddOp):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a + b

    @visitor.when(naturals_values.SubOp)
    def visit(self, node: naturals_values.SubOp):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a - b

    @visitor.when(naturals_values.MultiOp)
    def visit(self, node: naturals_values.MultiOp):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a * b

    @visitor.when(naturals_values.DivOp)
    def visit(self, node: naturals_values.DivOp):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a / b

    #################################################################
    #                                                               #
    #                  Logic Visit                                  #
    #                                                               #
    #################################################################

    @visitor.when(naturals_values.AndOp)
    def visit(self, node: naturals_values.AndOp):
        a = self.visit(node.a)

        if not a:
            return True

        b = self.visit(node.b)

        return a and b

    @visitor.when(naturals_values.OrOp)
    def visit(self, node: naturals_values.OrOp):
        a = self.visit(node.a)

        if a:
            return True

        b = self.visit(node.b)

        return a or b

    #################################################################
    #                                                               #
    #                  Binary Cmp Visit                             #
    #                                                               #
    #################################################################
    @visitor.when(naturals_values.LessOp)
    def visit(self, node: naturals_values):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a < b

    @visitor.when(naturals_values.LessOrEqualOp)
    def visit(self, node: naturals_values.LessOrEqualOp):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a <= b

    @visitor.when(naturals_values.GreatOp)
    def visit(self, node: naturals_values.GreatOp):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a > b

    @visitor.when(naturals_values.GreatOrEqualOp)
    def visit(self, node: naturals_values.GreatOrEqualOp):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a >= b

    @visitor.when(naturals_values.NotEqualOp)
    def visit(self, node: naturals_values.NotEqualOp):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a != b

    @visitor.when(naturals_values.EqualOp)
    def visit(self, node: naturals_values.EqualOp):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a == b

    #################################################################
    #                                                               #
    #                  Simple Transform                             #
    #                                                               #
    #################################################################

    @visitor.when(naturals_values.GetAttr)
    def visit(self, node: naturals_values.GetAttr):
        a = self.visit(node.a)
        b = self.visit(node.b)

        value = getattr(a, b)
        try:
            _ = value.get_sample
        except AttributeError:
            return value

        return value.get_sample(a.__instance_context__)[0]

    @ visitor.when(naturals_values.GetItem)
    def visit(self, node: naturals_values.GetItem):
        a = self.visit(node.a)
        b = self.visit(node.b)

        return a[b] if type(b) != bool else a

    @ visitor.when(naturals_values.SpaceSelfNode)
    def visit(self, node: naturals_values.SpaceSelfNode):
        return node.a.get_sample(self.context)[0]

    @ visitor.when(naturals_values.NaturalValue)
    def visit(self, node: naturals_values.NaturalValue):
        return node.a

    @ visitor.when(naturals_values.IndexSelf)
    def visit(self, node: naturals_values.IndexSelf):
        try:
            return self.current_index[node.a]
        except IndexError:
            return True



from typing import Protocol


from search_space.context_manager import SamplerContext


class SearchSpaceProtocol(Protocol):
    def get_sample(self, context: SamplerContext = None, local_domain=None):
        """
        """



from .__base__ import check_natural_value
from search_space.utils.ast_tools import decorated_all_methods, check_ast_precedence
from .logic_ops import LogicOpsNode


@decorated_all_methods(check_ast_precedence)
@decorated_all_methods(check_natural_value)
class ExprOpsNode(LogicOpsNode):

    #################################################################
    #                                                               #
    #              Binary Arithmetic Operations                     #
    #                                                               #
    #################################################################

    def __add__(self, other):
        return AddOp(self, other)

    def __radd__(self, other):
        return AddOp(other, self)

    def __mul__(self, other):
        return MultiOp(self, other)

    def __rmul__(self, other):
        return MultiOp(other, self)

    def __sub__(self, other):
        return SubOp(self, other)

    def __rsub__(self, other):
        return SubOp(other, self)

    def __truediv__(self, other):
        return DivOp(self, other)

    def __rtruediv__(self, other):
        return DivOp(other, self)

    def __mod__(self, other):
        return ModOp(self, other)

    def __rdivmod__(self, other):
        return ModOp(other, self)

    #################################################################
    #                                                               #
    #                  Binary Cmp Visit                             #
    #                                                               #
    #################################################################

    def __eq__(self, other):
        return EqualOp(self, other)

    def __req__(self, other):
        return EqualOp(other, self)

    def __ne__(self, other):
        return NotEqualOp(self, other)

    def __rne__(self, other):
        return NotEqualOp(other, self)

    def __lt__(self, other):
        return LessOp(self, other)

    def __rlt__(self, other):
        return LessOp(other, self)

    def __gt__(self, other):
        return GreatOp(self, other)

    def __rgt__(self, other):
        return GreatOp(other, self)

    def __ge__(self, other):
        return GreatOrEqualOp(self, other)

    def __rge__(self, other):
        return GreatOrEqualOp(other, self)

    def __le__(self, other):
        return LessOrEqualOp(self, other)

    def __rle__(self, other):
        return LessOrEqualOp(other, self)

    #################################################################
    #                                                               #
    #              Binary Arithmetic Operations                     #
    #                                                               #
    #################################################################


class AddOp(ExprOpsNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.a).__name__} + ( {type(self.b).__name__} {op} ...'


class SubOp(ExprOpsNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.a).__name__} + ( {type(self.b).__name__} {op} ...'


class MultiOp(ExprOpsNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.a).__name__} + ( {type(self.b).__name__} {op} ...'


class DivOp(ExprOpsNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.a).__name__} + ( {type(self.b).__name__} {op} ...'


class ModOp(ExprOpsNode):
    def _suggestion(self, op):
        return f'try to change the constraint like \
            {type(self.a).__name__} + ( {type(self.b).__name__} {op} ...'

    #################################################################
    #                                                               #
    #                  Binary Cmp Visit                             #
    #                                                               #
    #################################################################


class EqualOp(ExprOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
                    {type(self.a).__name__} + ( {type(self.b).__name__} {op} ..."""


class NotEqualOp(ExprOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
                    {type(self.a).__name__} + ( {type(self.b).__name__} {op} ..."""


class GreatOrEqualOp(ExprOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
                    {type(self.a).__name__} + ( {type(self.b).__name__} {op} ..."""


class GreatOp(ExprOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
                    {type(self.a).__name__} + ( {type(self.b).__name__} {op} ..."""


class LessOrEqualOp(ExprOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
                    {type(self.a).__name__} + ( {type(self.b).__name__} {op} ..."""


class LessOp(ExprOpsNode):
    def _suggestion(self, op):
        return f"""try to change the constraint like 
                    {type(self.a).__name__} + ( {type(self.b).__name__} {op} ..."""


class IndexSelf(ExprOpsNode):
    def __init__(self, a) -> None:
        super().__init__(a, None)



from random import uniform
from search_space.sampler import distribution, Sampler
from search_space.sampler.distribution_names import UNIFORM


@distribution(UNIFORM)
class UniformSampler(Sampler):
    def generate_random_value(self, domain):
        a, b = domain
        return uniform(a, b)



from threading import currentThread
from search_space.errors import CircularDependencyDetected, DetectedRuntimeDependency, InvalidSpaceConstraint, NotEvaluateError
from ..asts import constraints
from search_space.utils import visitor
from search_space.spaces.asts import constraints, naturals_values
from . import VisitorLayer
from .visitor_natural_ast import NaturalAstVisitor, NaturalValuesNode


class IndexAstModifierVisitor(VisitorLayer):
    def __init__(self, space, index) -> None:
        self.current_index = index
        self.space = space
        self.natural_visitor = NaturalAstVisitor()

    def domain_optimization(self, node, domain):
        self.context = None
        return self.visit(node), domain

    def transform_to_modifier(self, node, domain=None, context=None):
        self.context = context
        return self.visit(node), domain

    def transform_to_check_sample(self, node, sample, context=None):
        self.context = context
        return self.visit(node, context=context)

    @visitor.on("node")
    def visit(self, node, current_index=[]):
        pass

    @visitor.when(constraints.AstRoot)
    def visit(self, node, current_index=[]):
        result = constraints.AstRoot([])

        for n in node.asts:
            result.add_constraint(self.visit(n, None))

        return result

    #################################################################
    #                                                               #
    #                  Binary Cmp Visit                             #
    #                                                               #
    #################################################################

    @visitor.when(constraints.UniversalVariableBinaryOperation)
    def visit(self, node, current_index):
        a = self.visit(node.target, current_index)
        b = self.visit(node.other, current_index)

        if isinstance(node, constraints.SegmentationExprNode):
            c = self.visit(node.value, current_index)
            return type.__call__(node.__class__, a, b, c)

        return type.__call__(node.__class__, a, b)

    #################################################################
    #                                                               #
    #                  Simple Transform                             #
    #                                                               #
    #################################################################

    @visitor.when(constraints.GetItem)
    def visit(self, node, current_index):
        current_index = [] if current_index is None else current_index
        _slice = node.other.target

        if isinstance(_slice, slice):
            start, stop, steep = (
                self.visit(
                    constraints.NaturalValue(_slice.start),
                    current_index
                ),
                self.visit(
                    constraints.NaturalValue(_slice.stop),
                    current_index
                ),
                self.visit(
                    constraints.NaturalValue(_slice.step),
                    current_index
                )
            )
            try:
                if bool in [type(start.target), type(stop.target), type(steep.target)]:
                    raise InvalidSpaceConstraint(
                        'the slice dont support dynamic indexes')

                index = constraints.NaturalValue(
                    slice(start.target, stop.target, steep.target)
                )
            except AttributeError:
                return constraints.NotEvaluate()
        else:
            index = self.visit(node.other, current_index)
            try:
                if type(index.target) == bool:
                    index.target = self.current_index[-len(current_index) - 1]
            except AttributeError:
                return constraints.NotEvaluate()

        return self.visit(node.target, [index.target] + current_index)

    @visitor.when(constraints.SelfNode)
    def visit(self, node, current_index):
        if current_index is None:
            return node

        if len(current_index) == 0:
            return constraints.NotEvaluate()

        if tuple(current_index) == self.current_index:
            return node

        try:
            if self.context is None:
                return constraints.NaturalValue(
                    naturals_values.SpaceSelfNode(self.space[current_index])
                )

            value = self.space[current_index].get_sample(
                context=self.context)[0]
        except NotEvaluateError:
            return constraints.NotEvaluate()
        except CircularDependencyDetected:
            return constraints.NotEvaluate()

        return constraints.NaturalValue(value)

    @visitor.when(constraints.NaturalValue)
    def visit(self, node, current_index):

        if isinstance(node.target, NaturalValuesNode):
            try:
                result = self.natural_visitor.get_value(
                    node.target,
                    context=self.context,
                    current_index=self.current_index
                )
                return constraints.NaturalValue(result)
            except CircularDependencyDetected:
                return constraints.NotEvaluate()

        return node

    @visitor.when(constraints.FunctionNode)
    def visit(self, node: constraints.FunctionNode, current_index):
        new_args = []
        for arg in node.args:
            new_args.append(self.visit(arg, current_index))

        new_kw = {}
        for name, arg in node.kwargs:
            new_kw[name] = self.visit(arg, current_index)

        return constraints.FunctionNode(node.func, new_args, new_kw)

    @visitor.when(constraints.AdvancedFunctionNode)
    def visit(self, node: constraints.AdvancedFunctionNode, current_index):
        new_args = []
        for arg in node.args:
            new_args.append(self.visit(arg, current_index))

        new_kw = {}
        for name, arg in node.kwargs:
            new_kw[name] = self.visit(arg, current_index)

        return constraints.AdvancedFunctionNode(
            (node.args_target, node.kw_target),
            node.func, new_args, new_kw,
            node.cls
        )



from search_space.utils.singleton import Singleton
from search_space.utils.infinity import Infinity


class Sampler:
    def __init__(self) -> None:
        self.history = {}

    def _get_true_range(self, a, b):
        if not Infinity in [type(a), type(b)]:
            return a, b

        if type(a) == Infinity and type(b) == Infinity:
            return float(a), float(b)

        if type(a) != Infinity:
            m = float(b)
            return (a, m) if a < m else (a, a*2)

        if type(b) != Infinity:
            m = float(a)
            return (m, b) if (m < b and abs(m) - abs(b) > 1) else (- abs(2*b), b)

    def generate_random_value(self, domain):
        pass

    def __random_value__(self, domain, repeat_last_value):
        if repeat_last_value:
            return self.history[domain][-1]

        return self.__add_to_history__(domain, self.generate_random_value(domain))

    def get_int(self, min, max, repeat_last_value=False):
        value = self.__random_value__(
            self._get_true_range(min, max), repeat_last_value=repeat_last_value)
        decimal = value - int(value)
        return int(value) if decimal <= 0.5 else int(value) + 1

    def get_float(self, min, max, repeat_last_value=False):
        return self.__random_value__(self._get_true_range(min, max), repeat_last_value=repeat_last_value)

    def choice(self, domain, repeat_last_value=False):
        index = self.get_int(
            0, len(domain) - 1, repeat_last_value=repeat_last_value)

        return domain[index]

    def get_boolean(self, domain=(0, 1), repeat_last_value=False):
        value = self.__random_value__(
            domain, repeat_last_value=repeat_last_value)
        return True if value > 0.5 else False

    def __add_to_history__(self, domain, value):
        try:
            self.history[domain].append(value)
        except KeyError:
            self.history[domain] = [value]
        return value

    def last_value(self, domain):
        if type(domain) == type(list()):
            domain = (0, len(domain) - 1)

        return self.history[domain][-1]


class SamplerDataBase(metaclass=Singleton):
    def __init__(self) -> None:
        self.db = {}

    def registry_sampler(self, name, cclass):
        self.db[name] = cclass

    def get_sampler(self, distribute_name):
        try:
            return self.db[distribute_name]()
        except:
            return None


class SamplerFactory:
    def __init__(self) -> None:
        self.db = SamplerDataBase()

    def create_sampler(self, distribute_name, search_space=None):
        return self.db.get_sampler(distribute_name)


class SamplerFactoryWithMemory(SamplerFactory):
    def __init__(self) -> None:
        super().__init__()
        self.__instances = {}

    def create_sampler(self, distribute_name, search_space=None):
        if distribute_name is None:
            return None

        if search_space is None:
            return self.db.get_sampler(distribute_name)

        try:
            return self.__instances[(search_space, distribute_name)]
        except KeyError:
            self.__instances[(search_space, distribute_name)
                             ] = self.db.get_sampler(distribute_name)

            return self.__instances[(search_space, distribute_name)]

    def refresh_all_samplers(self):
        for sample in self.__instances.values():
            sample._refresh()


def distribution(name=''):
    def f(classs):
        s = SamplerDataBase()
        s.registry_sampler(name, classs)
        classs.__distribute_name__ = name
        return classs

    return f



from unittest import result
from search_space.context_manager.sampler_context import SamplerContext
from search_space.errors import DetectedRuntimeDependency, InvalidSampler, InvalidSpaceConstraint, NotEvaluateError
from search_space.spaces import BasicSearchSpace
from copy import copy
from search_space.spaces.visitors import visitors
from search_space.spaces.asts import constraints as ast_constraint
# TODO: check space types
import numpy as np


class ListSlicePointer:
    def __init__(self, _slice, tensor, dims) -> None:
        self.slice: slice = _slice[0]
        self.tensor: ListSearchSpace = tensor
        self.dims = dims

    def get_sample(self, context: SamplerContext = None, local_domain=None):
        dynamic_result = context.get_sampler_value(self.tensor)

        for index in range(*self.slice.indices(self.dims)):
            if np.isnan(dynamic_result[index]):
                dynamic_result[index], context = self.tensor.__sample_index__(
                    (index,), context, local_domain)

        return dynamic_result[self.slice], context


class TensorSlicePointer:

    def __init__(self, index, tensor, dims) -> None:
        self.index = index
        self.tensor: TensorSearchSpace = tensor
        self.dims = dims

    def get_sample(self, context=None, local_domain=None):
        if context is None:
            context = SamplerContext(name="TensorSlicePointer")
        else:
            cache_value = context.get_sampler_value(self)
            if not cache_value is None:
                return cache_value, context

        sample = self._sample(context, local_domain, self.index)
        context.registry_sampler(self, sample)

        return sample, context

    def _sample(self, context, local_domain, index):
        if type(index) == type(tuple()):
            return self.tensor.samplers[index].get_sample(context, local_domain)[0]

        result = []
        for item in index:
            result.append(self._sample(context, local_domain, item))

        return result

    def __hash__(self) -> int:
        return id(self)


class TensorIndexPointer:

    def __init__(self, index, tensor) -> None:
        self.index = index
        self.tensor: TensorSearchSpace = tensor

    def get_sample(self, context=None, local_domain=None):
        return self.tensor.__sample_index__(self.index, context, local_domain)


class ListSearchSpace(BasicSearchSpace):

    #################################################################
    #                                                               #
    #                     Space Initialize                          #
    #                                                               #
    #################################################################

    def __init__(self, shape_space: list, distribute_like=None) -> None:
        super().__init__((), None)
        self.len_spaces = shape_space if type(
            shape_space) in [type(list()), type(tuple())] else [shape_space]

        self.samplers = {}
        self.ast_constraint = ast_constraint.AstRoot([])
        self.visitor_layers = []

    def set_type(self, space: BasicSearchSpace):
        if len(self.len_spaces) > 1:
            return TensorSearchSpace(self.len_spaces).set_type(space)

        self.type_space = space

        if isinstance(self.len_spaces[0], BasicSearchSpace):
            return self

        self._current_shape = self.len_spaces[0]
        for index in range(self.len_spaces[0]):
            self._create_sampler_by_index((index,))

        return self

    def _get_range_iter(self, limit, length):
        if isinstance(limit, int):
            return range(limit)

        return range(*limit.indices(length))

    def __copy__(self):
        result = super().__copy__()
        result.len_spaces = self.len_spaces
        result.type_space = self.type_space
        result.ast_constraint = self.ast_constraint

        for key, sampler in self.samplers.items():
            result.samplers[key] = copy(sampler)

        return result

    # def bfs_dimensions_computing(self, dims):
    #     Q = [(i,) for i in self._get_range_iter(dims[0], self.len_spaces[0])]

    #     for i in range(1, len(dims)):
    #         temp = []
    #         for j in self._get_range_iter(dims[i], self.len_spaces[i]):
    #             for n in Q:
    #                 temp.append(n + (j,))
    #         Q = temp

    #     return Q

    #################################################################
    #                                                               #
    #                     GetItem                                   #
    #                                                               #
    #################################################################

    def _check_limits(self, index):
        if len(index) != 1:
            raise InvalidSampler(
                f"shape error, the index {index} in dimension {self.len_spaces}")
        i = index[0]
        shape = self._current_shape
        try:
            if i < 0 or i >= shape:
                raise NotEvaluateError()
        except TypeError:
            if i.start < 0 or i.stop >= shape:
                raise NotEvaluateError()

    def _create_sampler_by_index(self, index):
        try:
            self.samplers[index]
        except KeyError:
            self.samplers[index] = copy(self.type_space)
            self.samplers[index].space_name = f'{self.samplers[index].space_name}_{index}'
            self.samplers[index].layers_append(*self.visitor_layers)
            self.samplers[index].layers_append(
                visitors.EvalAstChecked(),
                visitors.IndexAstModifierVisitor(self, index)
            )

            self.samplers[index] = self.samplers[index].__ast_optimization__(
                self.ast_constraint)

    def __getitem__(self, index):
        self._check_limits(index)
        if type(index) == type(list()):
            index = tuple(index)

        for i in index:
            if isinstance(i, slice):
                break
        else:
            self._create_sampler_by_index(index)
            return TensorIndexPointer(index, self)

        return ListSlicePointer(index, self, self._current_shape)

    #################################################################
    #                                                               #
    #                     Ast Optimization                          #
    #                                                               #
    #################################################################

    def __domain_optimization__(self, domain, ast_result):
        raise DetectedRuntimeDependency()

    def __advance_space__(self, ast: ast_constraint.AstRoot):
        self.ast_constraint.asts += ast.asts
        return self

    def __ast_optimization__(self, ast_list):

        super().__ast_optimization__(ast_list)

        for key, space in self.samplers.items():
            self.samplers[key] = space.__ast_optimization__(ast_list)

        return self

    def layers_append(self, *args):
        self.visitor_layers += list(args)
        for key in self.samplers.keys():
            self.samplers[key].layers_append(*args)

    #################################################################
    #                                                               #
    #                     Sample Generate                           #
    #                                                               #
    #################################################################

    def __sampler__(self, domain, context: SamplerContext):

        try:
            self._current_shape = self.len_spaces[0].get_sample(context)[0]
        except AttributeError:
            self._current_shape = self.len_spaces[0]

        result = [np.nan] * self._current_shape
        context = context.create_child(f'{self.space_name}_index')
        context.registry_sampler(self, result)
        for index in range(self._current_shape):
            index = index,
            self._create_sampler_by_index(index)
            result[index[0]] = self.samplers[index].get_sample(context)[0]

        return result, context

    def __sample_index__(self, index, context: SamplerContext, domain):
        dynamic_result = context.get_sampler_value(self)
        dynamic_result[index[0]], context = self.samplers[index].get_sample(
            context)
        return dynamic_result[index[0]], context

    def __domain_filter__(self, domain, context):
        return domain, self.ast_constraint

    # TODO: Check list

    def __check_sample__(self, sample, ast_result, context):
        return sample
        # self.iter_virtual_list(
        #     self._current_shape, [],
        #     lambda index: visitors.ValidateSampler().check_sample_by_index(
        #         ast_result, sample, context, index
        #     )
        # )


class TensorSearchSpace(BasicSearchSpace):

    #################################################################
    #                                                               #
    #                     Space Initialize                          #
    #                                                               #
    #################################################################

    def __init__(self, shape_space: list, distribute_like=None) -> None:
        super().__init__((), None)
        self.len_spaces = shape_space if type(
            shape_space) in [type(list()), type(tuple())] else [shape_space]

        self.samplers = {}
        self.ast_constraint = ast_constraint.AstRoot([])
        self._current_shape = []
        self.visitor_layers = []

    def set_type(self, space: BasicSearchSpace):
        self.type_space = space

        for size in self.len_spaces:
            if isinstance(size, BasicSearchSpace):
                break
        else:

            self._current_shape = self.len_spaces
            self.iter_virtual_list(
                self.len_spaces, [], lambda index: self[index])

        return self

    def __copy__(self):
        result = super().__copy__()
        result.len_spaces = []
        for item in self.len_spaces:
            _copy = copy(item)
            try:
                _copy.set_hash(hash(item))
            except AttributeError:
                pass
            result.len_spaces.append(_copy)

        result.type_space = self.type_space
        result.ast_constraint = self.ast_constraint

        for key, sampler in self.samplers.items():
            result.samplers[key] = copy(sampler)

        return result

    def iter_virtual_list(self, shape, index, func):
        if len(shape) == 0:
            return func(index)

        result = []
        for i in range(shape[0]):
            result.append(self.iter_virtual_list(shape[1:], index + [i], func))

        return result

    def _check_limits(self, index):
        if len(index) != len(self._current_shape):
            raise InvalidSampler(
                f"shape error, the index {index} in dimension {self._current_shape}")

        for i, shape in zip(index, self._current_shape):
            try:
                if i < 0 or i >= shape:
                    raise NotEvaluateError()
            except TypeError:
                if i.start < 0 or i.stop >= shape:
                    raise NotEvaluateError()

    def _create_sampler_by_index(self, index):
        try:
            self.samplers[index]
        except KeyError:
            self.samplers[index] = copy(self.type_space)
            self.samplers[index].space_name = f'{self.samplers[index].space_name}_{index}'
            self.samplers[index].layers_append(*self.visitor_layers)
            self.samplers[index].layers_append(
                visitors.EvalAstChecked(),
                visitors.IndexAstModifierVisitor(self, index)
            )

            self.samplers[index] = self.samplers[index].__ast_optimization__(
                self.ast_constraint)

    def _flatter_index(self, index, i=0, result=[]):
        if i >= len(index):
            result = tuple(result)
            self._create_sampler_by_index(result)
            return result

        if isinstance(index[i], int):
            return self._flatter_index(index, i + 1, result + tuple(index[i]))

        response = []
        for j in range(*index[i].indices(self._current_shape[i])):
            response.append(self._flatter_index(index, i + 1, result + [j]))

        return response

    def __getitem__(self, index):
        self._check_limits(index)
        if type(index) == type(list()):
            index = tuple(index)

        for i in index:
            if isinstance(i, slice):
                break
        else:
            self._create_sampler_by_index(index)
            return TensorIndexPointer(index, self)

        flatter_index = self._flatter_index(index)
        return TensorSlicePointer(flatter_index, self, self._current_shape)

    #################################################################
    #                                                               #
    #                     Ast Optimization                          #
    #                                                               #
    #################################################################

    def __domain_optimization__(self, domain, ast_result):
        raise DetectedRuntimeDependency()

    def __advance_space__(self, ast: ast_constraint.AstRoot):
        self.ast_constraint.asts += ast.asts
        return self

    def __ast_optimization__(self, ast_list):

        super().__ast_optimization__(ast_list)

        for key, space in self.samplers.items():
            self.samplers[key] = space.__ast_optimization__(ast_list)

        return self

    def layers_append(self, *args):
        self.visitor_layers += list(args)
        for key in self.samplers.keys():
            self.samplers[key].layers_append(*args)

    #################################################################
    #                                                               #
    #                     Sample Generate                           #
    #                                                               #
    #################################################################

    def __sampler__(self, domain, context: SamplerContext):

        shape = []
        for ls in self.len_spaces:
            try:
                shape.append(ls.get_sample(context)[0])
            except AttributeError:
                shape.append(ls)

        self._current_shape = tuple(shape)
        context = context.create_child(f'{self.space_name}_index')
        return self.iter_virtual_list(
            shape, [], lambda index: self[index].get_sample(context)[0]), context

    def __sample_index__(self, index, context: SamplerContext, domain):
        return self.samplers[index].get_sample(context, domain)

    def __domain_filter__(self, domain, context):
        return domain, self.ast_constraint

    # TODO: Check list

    def __check_sample__(self, sample, ast_result, context):
        return sample
        # self.iter_virtual_list(
        #     self._current_shape, [],
        #     lambda index: visitors.ValidateSampler().check_sample_by_index(
        #         ast_result, sample, context, index
        #     )
        # )


# TODO copy



import inspect

__all__ = ['on', 'when']


def on(param_name):
    def f(fn):
        dispatcher = Dispatcher(param_name, fn)
        return dispatcher
    return f


def when(param_type, frame=None):
    def f(fn):
        nonlocal frame
        frame = inspect.currentframe().f_back if frame is None else frame
        func_name = fn.func_name if 'func_name' in dir(fn) else fn.__name__
        dispatcher = frame.f_locals[func_name]
        if not isinstance(dispatcher, Dispatcher):
            dispatcher = dispatcher.dispatcher
        dispatcher.add_target(param_type, fn)

        def ff(*args, **kw):
            return dispatcher(*args, **kw)
        ff.dispatcher = dispatcher
        return ff
    return f


class WithoutDispatcher:
    pass


without_dispatcher = WithoutDispatcher()


class Dispatcher(object):
    def __init__(self, param_name, fn):
        # frame = inspect.currentframe().f_back.f_back
        # top_level = frame.f_locals == frame.f_globals
        self.param_index = self.__argspec(fn).args.index(param_name)
        self.param_name = param_name
        self.targets = {}

    def __call__(self, *args, **kw):
        typ = args[self.param_index].__class__
        d = self.targets.get(typ)
        if d is not None:
            return d(*args, **kw)
        else:
            issub = issubclass
            t = self.targets
            ks = t.keys()
            ans = [t[k](*args, **kw) for k in ks if issub(typ, k)]
            if len(ans) == 1:
                return ans.pop()
            return ans if len(ans) > 0 else without_dispatcher

    def add_target(self, typ, target):
        self.targets[typ] = target

    @staticmethod
    def __argspec(fn):
        # Support for Python 3 type hints requires inspect.getfullargspec
        if hasattr(inspect, 'getfullargspec'):
            return inspect.getfullargspec(fn)
        else:
            return inspect.getargspec(fn)



class Singleton(type):
    _instances = None

    def __call__(cls, *args, **kwargs):
        if cls._instances is None:
            cls._instances = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances



from copy import copy
from search_space.errors import InvalidSpaceDefinition
from search_space.spaces.domains.bached_domain import BachedDomain
from search_space.spaces.domains.__base__ import Domain, NumeralDomain
from search_space.sampler import Sampler
from search_space.sampler.distribution_names import UNIFORM
from . import __namespace__ as nsp
from typing import Iterable
from search_space.utils.itertools import is_iterable


class LinearTransformedDomain(Domain, metaclass=nsp.LinealTransformed):
    def __init__(
        self,
        original_domain: NumeralDomain,
        transformer, inverse, independent_value,
    ) -> None:
        self.original_domain = copy(original_domain)
        self.transformer = transformer
        self.inverse = inverse
        self.ind_v = independent_value

        self.original_domain.min = self._transform_origin_to_new(
            self.original_domain.min)
        self.original_domain.max = self._transform_origin_to_new(
            self.original_domain.max)

    def is_invalid(self):
        return self.original_domain.is_invalid()

    @property
    def limits(self):
        return self.original_domain.limits

    def _transform_origin_to_new(self, value):
        return self.transformer(value - self.ind_v)

    def _transform_new_to_origin(self, value):
        return self.inverse(value) + self.ind_v

    def _transform_values(self, values):
        if not is_iterable(values):
            return self._transform_origin_to_new(values)

        return [self._transform_origin_to_new(item) for item in values]

    def get_sample(self, sampler: Sampler):
        return self._transform_new_to_origin(self.original_domain.get_sample(sampler))

    def __eq__(self, __o: object) -> bool:
        self.original_domain = self.original_domain == self._transform_values(
            __o)
        return self

    def __ne__(self, __o: object) -> bool:
        self.original_domain = self.original_domain != self._transform_values(
            __o)
        return self

    def __lt__(self, __o: object) -> bool:
        self.original_domain = self.original_domain < self._transform_values(
            __o)
        return self

    def __le__(self, __o: object) -> bool:
        self.original_domain = self.original_domain <= self._transform_values(
            __o)
        return self

    def __ge__(self, __o: object) -> bool:
        self.original_domain = self.original_domain >= self._transform_values(
            __o)
        return self

    def __gt__(self, __o: object) -> bool:
        self.original_domain = self.original_domain > self._transform_values(
            __o)
        return self



from typing import Protocol, Callable, Union
from search_space.utils.namespace_injection import InjectorMetaclass, New
from search_space.sampler import Sampler


class DomainProtocol:

    def get_sample(self, sampler: Sampler):
        pass

    def __copy__(self):
        pass


class NumeralDomainProtocol(DomainProtocol):
    min: int
    max: int


class NaturalDomain(NumeralDomainProtocol, InjectorMetaclass):
    def __init__(self, _min, _max) -> None:
        self.min = _min
        self.max = _max


class ContinuosDomain(NumeralDomainProtocol, InjectorMetaclass):
    def __init__(self, _min, _max) -> None:
        self.min = _min
        self.max = _max


class BachedDomain(DomainProtocol, InjectorMetaclass):
    def __init__(self, *domain: list[DomainProtocol]) -> None:
        pass


class LogBachedDomain(BachedDomain, InjectorMetaclass):
    def __init__(self, *domain: list[DomainProtocol]) -> None:
        pass


class LinealTransformed(DomainProtocol, InjectorMetaclass):
    def __init__(
        self,
        original_domain: NumeralDomainProtocol,
        transformer: Callable,
        inverse: Callable,
        independent_value: Union[int, float] = None
    ) -> None:
        pass



from .bached_domain import BachedDomain, LogBachedDomain
from .linear_transform_domain import LinearTransformedDomain
from .natural_domain import NaturalDomain
from .continuos_domain import ContinuosDomain
from .categorical_domain import CategoricalDomain


from . import __namespace__ as domain_namespace



from random import uniform
from search_space.sampler import Sampler, distribution
from search_space.sampler.distribution_names import UNIFORM_BERNOULLI


@distribution(UNIFORM_BERNOULLI)
class UniformBernoulliSampler(Sampler):
    def generate_random_value(self, domain):
        r = uniform(0, 1)
        return domain[int(r / (1/len(domain)))]
